/* $Id: hls.c 3194 2005-03-26 02:38:09Z bremner $ */

#include <stdio.h>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif


#include "basis_equation.h"
#include "boundary.h"
#include "combinat.h"
#include "common_startup.h"
#include "debug.h"
#include "globaldef.h"
#include "hls.h"
#include "hyperline.h"
#include "int_queue.h"
#include "nalloc.h"
#include "nodedef.h"
#include "oms_opt.h"
#include "output.h"
#include "profile.h"
#include "pluecker.h"
#include "ridge_graph.h"
#include "signed_set.h"
#include "startup.h"
#include "shutdown.h"

int oms_NodeTag, oms_GlobalTag;

static bool BoundaryCheck(na_node_p, na_node_p, array_index_t);
static bool hl_bdy_hook(na_node_p local, na_node_p global,
		 int32_t basis, int32_t value);
static void hls_success(na_node_p, na_node_p);

int main(int argc, char** argv){

  na_node_p local,global;
  params_p params;

  startopt_t startopt=SO_hyperline;

  if (OPT_use_profile) 
    prof_init();

  global=cs_init_global(argc,argv,startopt);

  local=cs_alloc_local(global,startopt);
  local=hls_init_local1(global,local);
  local=cs_init_local(global,local,startopt,hls_read_boundary_hook);
  if (!local) 
    shutdown(global,HERE,"inconsistent input");

  hls_init_local2(global,local);
  
  params=na_access(global,SLOT_params);
  
  output_init(params);

  TryBothSigns(local,global,params->r+1);

  if (OPT_use_profile)
    prof_dump(params->profile_level);
  output_finish(params);
  return 0;
}


/** 
 * This function is to be called _only_ from startup.
 * 
 * @param state 
 * @param global 
 * @param basis 
 * @param sigma 
 * 
 * @return 1 if OK
 */
bool SetAndConsequences(   na_node_p	global,
			   na_node_p	state,
			   array_index_t	basis, 
			   sign_t	sigma){

  if (basis==0){
    sarray_p chi=na_access(state,SLOT_chirotope);
    sa_init(chi,0,sigma);
    return BoundaryCheck(state,global,0);
  } else {
    return SetSign(state,global,basis,sigma);
  }

}

void TryBothSigns(na_node_p local,na_node_p global,  
		  array_index_t element){

  SaveState(local);
  (void)ProcessLine(local,global,0,element,1);
  RestoreState(local);

  SaveState(local);
  (void)ProcessLine(local,global,0,element,-1);
  RestoreState(local);
}

/* 
 * Try all possible places to insert @element@ into row @line@ of 
 * the hyperline arrangment
 *
 * The sign with which @element@ is inserted implicitly on the sign
 * of @chirotope[0]@
 */
static bool ProcessLine( na_node_p local, na_node_p global,
			  array_index_t line, array_index_t element, sign_t sign0){
  array_index_t gap;
  sarray_p chirotope;
  hyperline_cfg_p H;


  chirotope=na_access(local,SLOT_chirotope);
  H=na_access(local,SLOT_hyp);

  if (OPT_oms_debuglevel>=1){
    /* verify all the signs on the previous line are set */
    if (line>0) EnsureSet(H,chirotope,line-1);
  }

  err_debug(DB_BACKTRACK, ("ProcessLine","elem=%d line=%d sign0=%d #unset=%d",element,line,sign0,sa_zeros(chirotope)));


  if (OPT_hls_shortcut) {
    /* to make this check worthwhile on sparse examples, 
       it would need a more efficient implementation of 
       sa_zeros, e.g. save a zero count in the (n+1)st element.
    */
    if (sa_zeros(chirotope)==0){
      /* ok, no point in inserting more elements, all signs are 
	 determined */
      hls_success(global,local);
      return 0;
    }
  }
  if (line==0){
    array_index_t  beta=hl_index(H,hl_head(H,0));
    sign_t sigma=sa_read(chirotope,beta);
    


    if (sigma!=0 && sigma!=sign0){
      err_debug(DB_BACKTRACK,("ProcessLine","sign0 conflict %d != %d",sign0,sigma));
      ERR_RETURN(FALSE);
    } else {
      if (sigma==0){
	sa_set(chirotope,beta,sign0); /* choose sign of [lambda_0, n+1]  */
	if (!BoundaryCheck(local,global,beta)){
	  ERR_RETURN(FALSE);
	}
      }
    }

  }

  for (gap=hl_left(H,line); gap>=0; gap=hl_toright(H,line,gap)){

    SaveState(local);

    if (OPT_use_profile) prof_event(PROF_NODE);

    TryGap( local,global, line, gap ,element,sign0);
    RestoreState(local);
  }
  ERR_RETURN( TRUE );
}

/*
 * Try inserting element @element@ into row @line@ at position @gap@.
 * Look at all of the chirotope orientations (signs) implied.
 * If successful call @process_line(line+1)@
 * Note that @gap@ is a "pointer", i.e. cell index, _not_ a column
 */

extern void TryGap( na_node_p local, 
		    na_node_p global,
		    array_index_t line, array_index_t gap,array_index_t element,sign_t sign0){

  array_index_t i,oleft,oright;
  int passed;
  sarray_p chirotope;
  hl_cfg_p H;

  err_debug(3,("TryGap","%d %d %d",element,line,gap));

  chirotope=na_access(local,SLOT_chirotope);
  H=na_access(local,SLOT_hyp);

  oleft=hl_left(H,line);
  oright=hl_right(H,line);

  if (!hl_check_boundary(H,chirotope,line,gap))
    return;


  hl_set_left(H,line,gap);
  hl_set_right(H,line,gap);

  i=oleft;
  do {
    
    int new_sign;
    array_index_t bindex;

    passed= (hl_key(H,gap) >= hl_key(H,i)) ? 1 : -1;


    new_sign=  passed*hl_real_sign(H,i)* 
      hl_new_real_sign(H,line,chirotope)*
      hl_tau(H,i);
    bindex=hl_index(H,i);

    if (!SetSign(local,global,bindex,new_sign)){
      err_debug(4,("TryGap","SetSign chi[%d] -> %d failed",bindex,new_sign));
      return;
    }
    i=hl_rnext(H,i);
  } while (hl_key(H,i) <= hl_key(H,oright) &&
	   hl_key(H,oleft) < hl_key(H,i) /* check for wrap */);


  
  if (line+1 == hl_rows(H)){
    err_debug(2,(HERE,"End of table reached for element %d",element));
    if (hl_cols(H) >= hl_max_cols(H)-1){
      /* We have inserted all elements */
      hls_success(global,local);
    } else {
      if (hl_update(local,global,H,chirotope)){
	TryBothSigns(local,global,element+1);
      }
    }
    return;
  }

  /* otherwise */
  ProcessLine(local,global,line+1,element,sign0);
  return;
}

static void EnsureSet(hl_cfg_p H, sarray_p chirotope, array_index_t line){
  array_index_t i,head;

  head=hl_head(H,line);

  for (i=0 ; i< hl_cols(H); i++){
    sign_t sigma;
    array_index_t b;

    b=hl_index(H,head+i);
    sigma=sa_read(chirotope,b);

    if (sigma*sigma != 1){
      err_fatal("EnsureSet","unset index %d = %d in line %d",b,sigma,line);
    }
  }
}

static bool BoundaryCheck(na_node_p local, na_node_p global,
			  array_index_t bindex){
  signed_set_p S;
    sarray_p chirotope;
    bool consistent=TRUE;

    chirotope=na_access(local,SLOT_chirotope);
    S=ss_new(sa_size(chirotope));

    consistent=be_boundary_traverse2(bindex,local,global,S,hl_bdy_hook);


    while(consistent && !ss_empty(S)){
      int choice;
      array_index_t beta;
      sign_t sigma;

      
      choice=ss_deq(S);
      beta=si_index(choice);
      sigma=si_sign(choice);
      sa_set(chirotope,beta,sigma);


      consistent=hls_pluecker_traverse(beta,local,global,S);
      if (!consistent){
	err_debug(3,(HERE,"pluecker eqns failed"));
      }

      consistent=be_boundary_traverse2(beta,local,global,S,hl_bdy_hook);
      if (!consistent){
	err_debug(3,(HERE,"boundary eqns failed"));
      }


    }
    ss_destroy(S);
    return consistent;
}

extern bool SetSign(na_node_p local, na_node_p global,
		    array_index_t bindex, sign_t sign){

  sarray_p chirotope;
  int cur_sign;
 

  err_assert(sign*sign==1);

  err_debug(3,("SetSign","%d %d",bindex,sign));
  
  chirotope=na_access(local,SLOT_chirotope);
  cur_sign=sa_read(chirotope,bindex);
  
  if (cur_sign==sign) 
    ERR_RETURN(TRUE);

  if (cur_sign == 0){
    sa_set(chirotope,bindex,sign);
    return FindConsequences(local,global,bindex);
  }
  ERR_RETURN(FALSE);
}

bool FindConsequences(na_node_p local, na_node_p global,
		      array_index_t bindex){

  sarray_p chirotope;
  array_index_t table_beta;
  signed_set_p S;
  sign_t sign;
  hl_cfg_p H;
  bool rval;


  chirotope=na_access(local,SLOT_chirotope);
  sign=sa_read(chirotope,bindex);
  H=na_access(local,SLOT_hyp);
  S=ss_new(sa_size(chirotope));

  /* note, this could be made slightly more efficent by only checking 
     the first index, and not redundantly setting it 
  */
  ss_add(S,bindex,sign);
  

  rval=TRUE;
  while(rval && !ss_empty(S)){
    int choice;
    array_index_t beta;
    sign_t sigma,old_sign;

    choice=ss_deq(S);
    beta=si_index(choice);
    sigma=si_sign(choice);
    
    old_sign=sa_read(chirotope,beta);

    if (old_sign != sigma && old_sign !=0){
      rval=FALSE;
    } else {
      if (old_sign==0)	sa_set(chirotope,beta,sigma);
    
      rval=be_boundary_traverse2(beta,local,global,S,hl_bdy_hook);

      rval= rval && hls_pluecker_traverse(beta,local,global,S);
      
      if (H->inited && rval && beta!=0){
	table_beta=beta+(H->max_cols-H->cols-1);

	/* check to see if the elements for this bindex have 
	   been inserted yet 
	*/
	if (table_beta < sa_size(H->clist)){
	  rval=hl_set_gaps(local,global,chirotope,H,table_beta,sigma);
	  err_debug(4,("FindConsequences","set_gaps %d = %d",beta,rval));
	}
      }
    }
  }
  ss_destroy(S);
  ERR_RETURN(rval);

}

/* currently unused in hls */
void local_compact(na_node_p local){
}


static bool hl_bdy_hook(na_node_p local, na_node_p global, 
		 int32_t basis, int32_t value){

  if (value==1){
    hl_cfg_p H=na_access(local,SLOT_hyp);
    be_block_p ridgetab=na_access(global,SLOT_ridge_table);
    return hl_set_ridges(H,basis,ridgetab);
  }
  return 1;
}


static void hls_success(na_node_p global, na_node_p local){
  hl_cfg_p H=na_access(local,SLOT_hyp);
  params_p params=na_access(global,SLOT_params);  
  maybe_prof_event(PROF_SUCCESS);

  err_debug(DB_BACKTRACK,(HERE,"success!"));
  if (OPT_oms_debuglevel >=2 ){
    hl_dump(H);
  }

  output_state(global,local);
  if (params->max_results>=1 && params->max_results<=prof_get(PROF_SUCCESS)){
    prof_dump(params->profile_level);
    output_finish(params);
    exit(0);
  }

}

/** 
 * check to see if extra pluecker checks are enabled, and maybe call actual check
 *
 * 
 * @param bindex basis index
 * @param local search state
 * @param global global (readonly) data
 * @param S signed set of facets
 * 
 * @return true if either the checks are disabled, or the return true.
 */ 

bool hls_pluecker_traverse(array_index_t bindex, na_node_p local,
			   na_node_p global,signed_set_p S){
  
  params_p params=na_access(global,SLOT_params);

  if (params->extra_pluecker){
    prof_event(PROF_EXTRA_PLUECKER);
    return pl_traverse2(bindex,local,global,S);
  } else {
    return 1;
  }
}

static na_node_p hls_init_local1(na_node_p global,na_node_p local){

  params_p params=na_access(global,SLOT_params);
  int n=params->n;
  int r=params->r;

  local=hl_nalloc(local,SLOT_hyp,n,r);
  na_fixup(local);
  return local;
}


static void  hls_init_local2(na_node_p global,na_node_p local){

  params_p params=na_access(global,SLOT_params);
  int n=params->n;
  int r=params->r;
  sarray_p chirotope;
  sign_t sigma0;
  hl_cfg_p H=na_access(local,SLOT_hyp);

  chirotope=na_access(local,SLOT_chirotope);

  sigma0=sa_read(chirotope,0);

  if (sigma0==0){
    err_warning("hls","forcing basis 0 to +");
    if (!SetAndConsequences(global,local,0,1)){
      err_warning("hls","Unable to initialize basis 0");
      if (OPT_use_profile)
	prof_dump(params->profile_level);
      exit(0);
    }
    sigma0=1;
  } 
  hl_init(H,n,r,sigma0);
}


bool hls_read_boundary_hook(na_node_p global,na_node_p local, 
			      int index, int sigma){
  hl_cfg_p   H=na_access(local,SLOT_hyp);
  be_block_p ridge_table=na_access(global,SLOT_ridge_table);

  if (sigma>0) 
    return hl_set_ridges(H,index,ridge_table);
  else
    return TRUE;


}

void SaveState(na_node_p state){

  lparams_p lparams=na_access(state,SLOT_lparams);
  lparams->depth++;

  rg_save(na_access(state,SLOT_ridge_graph));
  hl_save(na_access(state,SLOT_hyp));

  sa_save(na_access(state,SLOT_pos_mask));
  sa_save(na_access(state,SLOT_neg_mask));
  sa_save(na_access(state,SLOT_chirotope));

  sa_save(na_access(state,SLOT_boundary));
}


extern void RestoreState(na_node_p state){

  lparams_p lparams;

  NODECHECK(state);

  lparams=na_access(state,SLOT_lparams);

  lparams->depth--;

   rg_restore(na_access(state,SLOT_ridge_graph));
   hl_restore(na_access(state,SLOT_hyp));

   sa_restore(na_access(state,SLOT_pos_mask));
   sa_restore(na_access(state,SLOT_neg_mask));
   sa_restore(na_access(state,SLOT_chirotope));
   sa_restore(na_access(state,SLOT_boundary));
 }
