/* %|$Id: hyperline.c 3182 2005-03-01 17:15:38Z bremner $|% */


/**
  \file hyperline.c
  \brief Functions for manipulating hyperline arrangments.
  
    Notation, terminology:

  Let \f$N\f$ denote the number of elements. Let \f$R=d+1\f$ denote the rank.  A
  (partial) hyperline configuration for \f$n \leq N\f$ elements is a
  sequence of rows 
  \f[H[i]=(\lambda_{i1}\ldots \lambda_{i(d-1)}\mid 
  s_{i1}\cdot\mu_{i1}\ldots s_{i(n-d+1)}\cdot\mu_{i(n-d+1)})\f]
  where \f$\lambda_{i} \not=  \lambda_{j}\f$, \f$i \not= j\f$
  and \f$\mu_{i}\f$ is a permutation of  \f$\{1\ldots{n}\} \setminus \lambda_{i}\f$ 
  and \f${s_{ij} \in \{ +1, -1 \}}\f$. Note: \f$\lambda_{i}\f$ is the ``hyperline''.

  Let \f$\chi\in \pm 1 ^{1..n \choose R}\f$.
  \f$\chi\f$ corresponds to a chirotope iff 
  \f$[ \lambda_{l},s_{ls}\cdot\mu_{ls}, s_{lt}\cdot\mu_{lt},]=1\f$ for 
  all \f$l\f$, \f$s<t\f$ for some
  hyperline configuration.
  
  It is sometimes useful to talk about a canonical bracket sign,
  i.e. \f$\langle \sigma_1 \mu_1 \sigma_2 \mu_2 \ldots \sigma_r \mu_r
  \rangle\f$ represents \f$\Pi_i \sigma_i [ m_1 m_2 \ldots m_k ]\f$ where 
  the \f$m\f$'s are the \f$\mu\f$'s in sorted order.
  

  Let \f$\tau(L)\f$ denote \f$(-1)^k\f$ where \f$k\f$ is the number of
  transpositions necessary to sort \f$L\f$. 

  Let \f$\beta(L,n)\f$ denote the index of \f$L\f$ in the canonical listing 
  of \f${1\ldots{}n \choose |L|}\f$.

  We store something slightly tricky  for the indices and signs.
  - \f$\mathtt{index}[i][j]=\beta(\{\lambda\sb{i},\mu\sb{ij}, N\}, N)\f$;
  i.e. we store the basis index with respect to the final element inserted.
  The index with respect to the current element can be obtained by
  subtraction.
  - \f$\mathtt{sign}[i][j]=s\sb{ij}\cdot \tau(\lambda\sb{i}, \mu\sb{ij})\f$ 
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include "oms_opt.h"
#include "oms_types.h"
#include "oms_error.h"
#include "output.h"
#include "combinat.h"
#include "statearray.h"
#include "hyperline.h"
#include "ridge_graph.h"
#include "int_queue.h"
#include "hl_internal.h"
#include "basis_equation.h"
#include "profile.h"

TAG_DECLARE(hl_fixup);

/*
 * delete the last inserted element from the hyperline config.
 */
static void hl_rollback(hl_cfg_p H){
  int i,n,r;

  r=H->rank;
  n=H->cols+r-2;
  H->rows -= binomial(n-1,r-3);

  H->cols--;
  for (i=0; i<H->rows; i++){
    array_index_t to_die,prev,next;
    to_die=hl_head(H,i)+H->cols;
    prev=hl_rprev(H,to_die);
    next=hl_rnext(H,to_die);

    H->cell[prev].rnext=next;
    H->cell[next].rprev=prev;
  }
}

extern bool hl_set_sign(hl_cfg_p H, 
			sarray_p chirotope,
			array_index_t bindex, sign_t sign){
  
  int cur_sign;
  array_index_t cur_cell;
  int num_occurances,i;
  array_index_t * occurance;

  oms_assert(sign*sign==1);

  oms_debug(3,("hl_set_sign","%d %d",bindex,sign));


  cur_sign=sa_read(chirotope,bindex);
  
  if (cur_sign == 0){
    sa_set(chirotope,bindex,sign);

    /* where else does this bracket occur? */

    /* the occurances must be treated from top of table. Possibly a
       change of data structure is in order, since the current one
       stores the from the bottom up */
    
    num_occurances=binomial(H->rank-1,H->rank-2);
    occurance=calloc(num_occurances,sizeof(array_index_t));

    i=num_occurances;
    cur_cell=sa_read(H->clist,hl_index_inv(H,bindex));
    while(cur_cell>=0){
      occurance[--i]=cur_cell;
      cur_cell=H->cell[cur_cell].cnext;
    }
    for (i=0; i< num_occurances; i++){

      hl_cell_p cell;

      cur_cell=occurance[i];
      cell=H->cell+cur_cell;

      if (hl_lfeasible(H,cur_cell)){
	/* some gap bounds have changed */

	if (  sign*hl_new_sign(H,cell->row,chirotope)*(cell->sign)==-1){
	  array_index_t k;

	  /* Right gap bound changed */

	  array_index_t old_right=hl_right(H,cell->row);
	  hl_set_right(H,cell->row,cell->rprev);

	  for (k=cell->rnext; 
		hl_key(H,cur_cell)< hl_key(H,k) &&hl_key(H,k) <= hl_key(H,old_right);
	       k=hl_rnext(H,k)){
	    /* these are the cocircuits we skipped over */
	    hl_cell_p ocell=H->cell+k;
	    
	    int new_sign=hl_new_sign(H,ocell->row,chirotope)*(ocell->sign)*-1;
	    array_index_t new_bindex=hl_index(H,k);


	    if (!hl_set_sign(H,chirotope,new_bindex,new_sign)){
	      free(occurance);
	      return FALSE;
	    }
	  }
	} else {
	  /* Left gap bound changed */
	  int k;
	  array_index_t old_left = hl_left(H,cell->row); 
    	  hl_set_left(H,cell->row,cur_cell);
    	  for ( k=hl_rnext(H,old_left); 
		hl_key(H,old_left)< hl_key(H,k) &&hl_key(H,k) < hl_key(H,cur_cell);
		k=hl_rnext(H,k) )  
	  { 
	    hl_cell_p ocell=H->cell+k;

            int new_sign  = hl_new_sign(H,ocell->row,chirotope) * ocell->sign;
            int new_bindex = hl_index(H,k);

    	    if ( !hl_set_sign(H,  chirotope, new_bindex, new_sign) ) {
	      free(occurance);
	      return FALSE;
	    }
    	  }
	}
      }
    } 
    free(occurance);
    return TRUE;
  } else {
    if (cur_sign==sign) 
      return TRUE;
  }

  return FALSE;
}

extern void hl_save(hl_cfg_p H){
  
  oms_assert(H->coltop < H->colstacksize);

  H->colstack[H->coltop++]=H->cols;

  sa_save(H->clist);
  sa_save(H->left);
  sa_save(H->right);
  sa_save(H->ridge);
}

extern void hl_restore(hl_cfg_p H){
  int old_cols;

  sa_restore(H->clist);
  sa_restore(H->left);
  sa_restore(H->right);
  sa_restore(H->ridge);


  old_cols=H->colstack[--(H->coltop)];

  while (H->cols > old_cols){
    hl_rollback(H);
  }
}

extern rank_t hl_key(hl_cfg_p H, array_index_t i){
  return H->cell[i].key;
}

/*
 * determine if the current cell is between the left bound and 
 * the right bound for the row that contains it, i.e. fake the 
 * test @hl_left(H,line) <= col   && col <= hl_right(H,line)@
 */

static bool hl_feasible(hl_cfg_p H,array_index_t i){
  hl_cell_p cell=H->cell;
  array_index_t l,r;
  l=hl_left(H,cell[i].row);
  r=hl_right(H,cell[i].row);

  return (cell[l].key <= cell[i].key   && cell[i].key<=cell[r].key );
}

/* as above, but insist on strict inequality on the left */
extern bool hl_lfeasible(hl_cfg_p H,array_index_t i){
  hl_cell_p cell=H->cell;
  array_index_t l,r;
  l=hl_left(H,cell[i].row);
  r=hl_right(H,cell[i].row);

  return (cell[l].key < cell[i].key   && cell[i].key<=cell[r].key );
}

/*%
  Compute the sign of $n+1$ when inserted into row @line@.
  The idea is that since we always insert after column 0, 
  $[ \lambda_{i} \; s_{i0}\mu_{i0}\; s_{n+1}(n+1)]=1$, 
  hence $s_{n+1}=\mathrm{sign}[i][0]\cdot \chi[\mathrm{index}[i][0]]$.
  NOTE: believe code, not comment
%*/

extern int hl_new_sign(hl_cfg_p H, array_index_t line,sarray_p chirotope){
  array_index_t k=hl_head(H,line);
  sign_t sigma;
  sigma=H->cell[k].sign*sa_read(chirotope,hl_index(H,k));
  return sigma;
}

/* Same thing, but with out basis sorting information */
extern int hl_new_real_sign(hl_cfg_p H, array_index_t line,sarray_p chirotope){
  array_index_t k=hl_head(H,line);
  sign_t sigma;
  /* We want %$\langle \lambda_i \mu_0 \sigma \mu_{n+1} \rangle=1$%,      so */
  sigma=H->cell[k].tau*sa_read(chirotope,hl_index(H,k)); 
  return sigma;
}

/* 
 * update the basis index table encoding the hyperline configuration 
 * after the gap bounds have been computed for element @n@
 *
 * Assumptions: @bd_left[i]@ contains the index  at which 
 * element @n@ should be inserted.
 *
 * Output: For hyperline row @H[i]@ @index_tab[i][j]@ contains 
 * @kset_index(n,d+1,{lambda_i,mu_ij,n+1}@.
 */

extern bool hl_update(na_node_p local, na_node_p global,
		      hl_cfg_p H,sarray_p chirotope){
  int i,n,r,N;

  bool rval;
  natural_t *pi,*pibar,*work;
  sign_t *tau;
  bool *queued;
  int_queue_p Q;



  r=H->rank;
  n=H->cols+r-1;
  N=H->max_cols+r-2;

  oms_assert(n<N);

  oms_debug(4,("hl_update","%d %d",n,N));


  /*
    see comments in combinat.c explaining k-partitions, 
    including the next three arrays
  */
  pi=calloc(r,sizeof(natural_t));
  pibar=calloc(n-r+2,sizeof(natural_t));
  tau=calloc(n-r+2,sizeof(sign_t));

  work=calloc(2*(n-r+2),sizeof(natural_t));

  queued=calloc(binomial(N,r),sizeof(bool));
  Q=iq_new(binomial(N,r));

  pi[0]=n;	 
  pi[1]=N;
  

  for (i=0; i<H->rows; i++){
    array_index_t k,beta,new_k;
    hl_cell_p cell,new_cell;
    unsigned long p_key,n_key;

    k=hl_left(H,i);
    oms_assert(k==hl_right(H,i));
    cell=H->cell+k;

    new_k=hl_head(H,i)+H->cols;
    new_cell=H->cell+(new_k);

    /* finish off our interrupted basis index calculation */

    oms_assert(H->mu_last[i]>0 && H->row_bindex[i]>=0);

    beta=kset_index_complete(N,r,pi,
			     r-2,r,0,H->mu_last[i],H->row_bindex[i]);

    new_cell->sign=hl_new_sign(H,i,chirotope);
    new_cell->real_sign=hl_new_real_sign(H,i,chirotope);
    new_cell->mu=n;
    new_cell->tau=1;  /* no sorting is necessary */
    new_cell->bindex=beta;
    new_cell->cnext=-1;
    hl_coeff_insert(H,new_k,beta);
    new_cell->rprev=k;
    new_cell->rnext=cell->rnext;
    new_cell->row=i;
    H->cell[cell->rnext].rprev=new_k;
    H->cell[k].rnext=new_k;

    p_key=hl_key(H,new_cell->rprev);
    n_key=hl_key(H,new_cell->rnext);
    
    if (n_key==0){
      n_key=KEY_MAX;
    }

    /* avoid silent (grr) integer overflows */
    new_cell->key=(p_key/2+n_key/2);  
    
    oms_assert( p_key < new_cell->key  && new_cell->key < n_key);

    /* Can we do something useful with alternation checking here?
     2002/01/07, db */
    hl_set_left(H,i,hl_head(H,i));
    hl_set_right(H,i,hl_rprev(H,hl_head(H,i)));
  }
  
  /* Now add the new rows */
  
  pi[r-1]=N;
  pi[r-3]=n;
  i=H->rows;
  rval=TRUE;
  first_kpartition(n-1,r-3,pi,pibar,tau);
  do {
    
    hl_cell_p cell;
    array_index_t j,head;
    head=hl_head(H,i);

    H->row_hlindex[i]=kset_cindex(N,r-2,pi);
    H->hlindex_row[H->row_hlindex[i]]=i;

    /* we leave the first one in place */
    hl_sort(N,r,pi,pibar[0],pibar+1,n-r+1,tau+1,chirotope,work);

    pi[r-1]=N;
    if (n <= N-2) {
      H->mu_last[i]=n;
      H->row_bindex[i]=kset_index_complete(N,r,pi,0,r-2,0,0,0);
    } else {
      H->mu_last[i]=-1;
      H->row_bindex[i]=-1;
    }

    for (j=0; j< n-r+2; j++){
      array_index_t beta,beta_0j,real_beta;

      cell=H->cell+head+j;
      cell->row=i;
      pi[r-1]=N;
      pi[r-2]=pibar[j];

      beta=kset_cindex(N,r,pi);
      cell->bindex=beta;
      cell->mu=pibar[j];	

      real_beta=beta-(N-n-1);
	
      if (!queued[real_beta]  && sa_read(chirotope,real_beta)!=0){
	oms_debug(4,("hl_update","queueing %d",real_beta));
	iq_enq(Q,real_beta);
	queued[real_beta]=TRUE;
      }

      /* this business of saving mu and tau is a late addition to 
	 try and make things simpler to understand.
	 probably there is now redundant cruft.

	 2001/11/05, db
      */

      if (j==0){
	/* one more flip to get over the new element */
	cell->sign=-1*tau[0];
	cell->tau=cell->sign;
	cell->real_sign=1;
      }else{
	int sigma_0j;

	pi[r-2]=pibar[0];
	pi[r-1]=pibar[j];
	beta_0j=kset_cindex(N,r,pi);

	sigma_0j=sa_read(chirotope,beta_0j);

	cell->sign=-1*tau[0]*sigma_0j;

	cell->tau=-1*tau[j];
	cell->real_sign=tau[0]*tau[j]*sigma_0j;
      }

      cell->rnext=head+((j+1) % (n-r+2));
      cell->rprev= (j==0) ? head+ (n-r+2-1) : head+j-1;

      cell->key=j*(KEY_MAX/(H->cols+1));

      cell->cnext= -1;
      hl_coeff_insert(H,head+j,beta);
    }
    /* The claim here is that the signed permutation produced by hl_sort
       is unique, so if it is alternating, then we should prune
       THIS CHECK IS CURRENTLY BOGUS
       because the value of H->cols  is not updated yet.
    */
    if (sa_read(H->ridge,H->row_hlindex[i]) &&  (!hl_check_alternation(H,i))){
      rval=FALSE;
    } else {
      hl_set_left(H,i,hl_head(H,i));
      hl_set_right(H,i,hl_rprev(H,hl_head(H,i)));

      i++;
    }
  } while (rval && next_kpartition(n-1,r-3,pi,pibar,tau));

  H->rows+=binomial(n-1,r-3);
  H->cols++;


  free(work);
  free(pi);
  free(pibar);
  free(tau);
  free(queued);

  while(rval==TRUE && !iq_empty(Q)){
    array_index_t k;

    k=iq_deq(Q);
    /* note, there is an extra sa_read here */

    rval=hl_set_gaps(local,global,chirotope,H,
		     k+(N-n-1),
		     sa_read(chirotope,k));
  }
  iq_destroy(Q);

  if (rval && OPT_oms_debuglevel>=1){
    for (i=0; i< H->rows; i++){
      if (sa_read(H->ridge,H->row_hlindex[i]) &&  (!hl_check_alternation(H,i))){
	hl_dump(H);
	oms_fatal("hl_update","ridge row %d alternates\n",i);
      }
    }
  }

#if OMS_DEBUGLEVEL>3
  hl_dump(H);
#endif

  RETURN(rval);
}

/*
  Check that an about to be  inserted hyperline element does not 
   create a sign alternation in a boundary row.
   PRECONDITION: If the row is boundary, then it was alternation 
   free before inserting this element.
   
   @row@     which row are we inserting into
   @k@       where in the row. this is a pseudopointer, i.e. cell index
	     we will insert after this spot.
*/

 int debug_count=0;

extern bool hl_check_boundary(hl_cfg_p H, sarray_p chirotope,
			      array_index_t row, array_index_t k){


  if (sa_read(H->ridge,H->row_hlindex[row])){
    int n_sign, p_sign, my_sign, head;
    hl_cell_p C=H->cell+k;

    oms_assert(hl_check_alternation(H,row));

    head=hl_head(H,row);
 

    /* here we want the sign of the next element, without the 
       sign of the first element, since otherwise this 
       will be included twice
    */
    my_sign=hl_new_real_sign(H,row,chirotope);
    p_sign=C->real_sign;

    n_sign=H->cell[C->rnext].real_sign;

    
    /* inserting at end? */
    if (C->rnext==head){
      if ((n_sign!=p_sign) && (p_sign != my_sign)){
	oms_debug(3,("hl_check_boundary",
		     "failing at end %d %d %c%c%c %s\n",
		     row,k, 
		     int_to_symbol(p_sign),
		     int_to_symbol(my_sign),
		     int_to_symbol(n_sign),
		     string_pm(chirotope)));
	if (OPT_oms_debuglevel>=3)
	  hl_dump(H);
	if (OPT_use_profile){
	  prof_event(PROF_RIDGE_CHECK);
	}
	RETURN(FALSE);
      } 
    }else {
      /* inserting in middle */
      if ((n_sign != my_sign ) && (p_sign != my_sign) ) {
	oms_debug(3,("hl_check_boundary","mid (%d): %d %d %c%c%c %s\n",
		     debug_count++,
		     row,k, 
		     int_to_symbol(p_sign),
		     int_to_symbol(my_sign),
		     int_to_symbol(n_sign),
		     string_pm(chirotope)));

	if (OPT_oms_debuglevel>=3){
	  hl_dump(H);
	}
	if (OPT_use_profile){
	  prof_event(PROF_RIDGE_CHECK);
	}


	/*	if (debug_count ==6) 
		return 1; */
	RETURN(FALSE); 
      }
    }
    oms_debug(3,("hl_check_boundary","OK: %d %d\n",row,k));
  } 



  return TRUE;
}

/* note, this could be made  more efficient */
static sign_t hl_cmp(int n, int r, array_index_t *pi,
		   natural_t  mu_0, natural_t mu_j, natural_t mu_k, 
		   sarray_p chirotope){
  array_index_t beta_jk,beta_0j,beta_0k;
  pi[r-2]=mu_0;
  pi[r-1]=mu_j;
  beta_0j=kset_cindex(n,r,pi);
  
  pi[r-1]=mu_k;
  beta_0k=kset_cindex(n,r,pi);

  pi[r-2]=mu_j;
  beta_jk=kset_cindex(n,r,pi);
  return sa_read(chirotope,beta_0j)*sa_read(chirotope,beta_0k)*
    sa_read(chirotope,beta_jk)  * (  mu_j   < mu_k ?  1 : -1);
}
    
    

static void hl_sort(int n, int r, array_index_t *pi,array_index_t  pi0, 
		    array_index_t *pibar, size_t len, 
		    array_index_t *tau,
		    sarray_p chirotope,
		    array_index_t *work){
  int j,j1,j2,s;
  size_t split;
  natural_t *work2=work+len;

  if (len==1) 
    return;

  split=len/2;

  hl_sort(n,r,pi,pi0,pibar,split,tau,chirotope,work);
  hl_sort(n,r,pi,pi0,pibar+split,len-split,tau+split,chirotope,work);
  
  j1=0; j2=split;

  for (j=0; j< len; j++){
    if (j1 >= split) 
      s=j2++;
    else if (j2 >=len) 
      s=j1++;
    else if  (hl_cmp(n,r,pi,pi0,pibar[j1],pibar[j2],chirotope) < 0 ) 
      s=j2++;
    else 
      s=j1++;

    work[j]=pibar[s];
    work2[j]=tau[s];
  }

  memcpy(pibar,work,len*sizeof(natural_t));
  memcpy(tau,work2,len*sizeof(sign_t));
}

extern array_index_t hl_rows(hl_cfg_p H){
  return H->rows;
}

extern array_index_t hl_cols(hl_cfg_p H){
  return H->cols;
}

extern array_index_t hl_max_cols(hl_cfg_p H){
  return H->max_cols;
}

extern array_index_t hl_sign(hl_cfg_p H, array_index_t i){
  return H->cell[i].sign;
}

extern array_index_t hl_tau(hl_cfg_p H, array_index_t i){
  return H->cell[i].tau;
}

extern array_index_t hl_real_sign(hl_cfg_p H, array_index_t i){
  return H->cell[i].real_sign;
}

/*return the @sign[line][col]@, where @sign[i][j]@%$=s\sb{ij}\cdot \tau(\lambda\sb{i}, \mu\sb{ij}) $%  */

/* 
 * compute the index with respect to the current element being inserted
 * 
 */
extern array_index_t hl_index(hl_cfg_p H,  array_index_t i){

  return H->cell[i].bindex-(H->max_cols-H->cols-1);
}

extern array_index_t hl_index_inv(hl_cfg_p H, array_index_t bindex){
  return H->bindex_table[bindex];
}

extern na_node_p hl_nalloc(na_node_p node, array_index_t slot, int n, int r){

  hl_cfg_p H;
  int rows,cells,cols,ops;
  array_index_t *bindex_table;

  rows=binomial(n,r-2);
  cols=n-r+2;
  cells=rows*cols;
  ops=binomial(n,r);
  

  node=na_alloc(node,slot,sizeof(hl_cfg_t)+
		sizeof(hl_cell_t)*cells);
  
  node=sa_nalloc(node,slot+HL_OFF_clist,binomial(n,r),ops);

  node=sa_nalloc(node,slot+HL_OFF_left,binomial(n,r-2),ops);

  node=sa_nalloc(node,slot+HL_OFF_right,binomial(n,r-2),ops);

  node=sa_nalloc(node,slot+HL_OFF_ridge,binomial(n,r-2),ops);
 
  node=na_alloc(node,slot+HL_OFF_row_bindex,binomial(n,r-2)*sizeof(natural_t));
  node=na_alloc(node,slot+HL_OFF_row_hlindex,binomial(n,r-2)*sizeof(natural_t));
  node=na_alloc(node,slot+HL_OFF_hlindex_row,binomial(n,r-2)*sizeof(natural_t));

  node=na_alloc(node,slot+HL_OFF_mu_last,binomial(n,r-2)*sizeof(natural_t));
  
  node=na_alloc(node,slot+HL_OFF_bindex_table,binomial(n,r)*sizeof(array_index_t));

  /** \todo this table is static, and should really be global */

  bindex_table=na_access(node,slot+HL_OFF_bindex_table);
  hl_init_bindex_table(bindex_table,n,r);

  node=na_alloc(node,slot+HL_OFF_colstack,
		(n+1)*binomial(n,r-2)*sizeof(natural_t));


  na_set_fixup(node,slot,TAG(hl_fixup));

  /* initialize as much as we can without knowing chirotope info*/

  
  hl_fixup(node,slot);
  H=na_access(node,slot);
  hl_preinit(H,n,r);

  return node;
						   
}

extern void hl_fixup(na_node_p node, array_index_t slot){
 hl_cfg_p H;
 H=na_access(node,slot);
 sa_fixup(node,slot+HL_OFF_clist);
 H->clist=na_access(node,slot+HL_OFF_clist);

 sa_fixup(node,slot+HL_OFF_left); 
 H->left=na_access(node,slot+HL_OFF_left);

 sa_fixup(node,slot+HL_OFF_right);
 H->right=na_access(node,slot+HL_OFF_right);

 sa_fixup(node,slot+HL_OFF_ridge);
 H->ridge=na_access(node,slot+HL_OFF_ridge);


 H->row_bindex=na_access(node,slot+HL_OFF_row_bindex);
 H->row_hlindex=na_access(node,slot+HL_OFF_row_hlindex);
 H->hlindex_row=na_access(node,slot+HL_OFF_hlindex_row);
 H->mu_last=na_access(node,slot+HL_OFF_mu_last);
 H->colstack=na_access(node,slot+HL_OFF_colstack);

 H->bindex_table=na_access(node,slot+HL_OFF_bindex_table);

}  


/** 
 * hl_preinit: carry out as much initialization as we can without knowing 
 * chirotope information
 * 
 * @param H hyperline structure
 * @param n number of elements
 * @param r rank
 */

static void hl_preinit(hl_cfg_p H,int n, int r){
  int i;


  H->colstacksize=(n+1)*binomial(n,r-2);
  H->inited=0;

  H->coltop=0;

  H->rank=r;
  H->max_cols=n-r+2;

  if ( ((rank_t)1<<(KEY_BITS-1))/(H->max_cols) <= (1<<(n-r)))
    oms_fatal("hl_init","internal hackery error. Chastise programmer");

  H->max_rows=binomial(n,r-2);
  H->rows=0;
  H->cols=2; 
  for (i=0; i< sa_size(H->clist); i++){
    sa_init(H->clist,i,-1);
  }

  for (i=0; i< sa_size(H->ridge); i++){
    sa_init(H->ridge,i,0);
  }

  for (i=0; i< binomial(n,r-2); i++){
    H->hlindex_row[i]=-1;
  }

  H->inited=0;
}


/* 
 * hl_init:  Finish initialization of  hyperline configuration.

 * H - hyperline arrangement
 * n - number of elements
 * r - rank
 * chi0 - sign of basis 0
 
*/
extern void hl_init(hl_cfg_p H,int n, int r,sign_t chi0){

  int i,line;
  natural_t *pi,pibar[2],tau[2];
 
  pi=calloc(r,sizeof(natural_t));
  TRAP_NULL(pi);

  pi[r-1]=n;

  line=0;
  first_kpartition(r,r-2,pi,pibar,tau);
  
  do {
    array_index_t k=hl_head(H,line);

    hl_set_left(H,line,k);
    hl_set_right(H,line,k+1);

    H->row_hlindex[line]=kset_cindex(n,r-2,pi);
    H->hlindex_row[H->row_hlindex[line]]=line;
    H->row_bindex[line]=kset_index_complete(n,r,pi,0,r-2,0,0,0);
    H->mu_last[line]=pi[r-3];

    for (i=0; i<2; i++){
      array_index_t beta;
      hl_cell_p cell=H->cell+k+i;

      cell->key=i*(KEY_MAX/2);
      cell->row=line;

      pi[r-2]=pibar[i];
      beta=kset_cindex(n,r,pi);
      cell->bindex=beta;
      cell->mu=pibar[i];
      cell->tau=tau[i];

      if (i==0) {
	cell->sign=tau[0];	
	cell->real_sign=1;
      } else {
	cell->sign=tau[0]*chi0;
	cell->real_sign=chi0*tau[0]*tau[i];
      }

      cell->rnext=k+(1-i);
      cell->rprev=k+(1-i);
      hl_coeff_insert(H,k+i,beta);
    }
    line++;
  } while (next_kpartition(r,r-2,pi,pibar,tau));
  H->rows=line;

  H->inited=1;
}
	      

extern array_index_t hl_left(hl_cfg_p H,array_index_t line){
  return sa_read(H->left,line);
}

extern array_index_t hl_right(hl_cfg_p H,array_index_t line){
  return sa_read(H->right,line);
}

extern void hl_set_right(hl_cfg_p H,array_index_t line,array_index_t val){
  oms_debug(6,("hl_set_right","%d %d->%d",line,sa_read(H->right,line),val));
   sa_set(H->right,line,val);
}

extern void hl_set_left(hl_cfg_p H,array_index_t line,array_index_t val){
  oms_debug(6,("hl_set_left","%d %d->%d",line,sa_read(H->left,line),val));
  sa_set(H->left,line,val);
}


extern array_index_t hl_toright(hl_cfg_p H,  array_index_t line, array_index_t k){
  array_index_t r;
  r=H->cell[k].rnext;
  if (r==hl_left(H,line) || !hl_feasible(H,r))
    return -1;
  else
    return r;
}


extern array_index_t hl_rnext(hl_cfg_p H,array_index_t k){
  return H->cell[k].rnext;
}

extern array_index_t hl_rprev(hl_cfg_p H,array_index_t k){
  return H->cell[k].rprev;
}

extern array_index_t hl_head(hl_cfg_p H,array_index_t line){
  return H->max_cols*line;
}


static void hl_coeff_insert(hl_cfg_p H, array_index_t where, array_index_t bindex){
  array_index_t head;

  head=sa_read(H->clist,bindex);
  
  H->cell[where].cnext=head;
  
  sa_set(H->clist,bindex,where);
}

extern void hl_dump_line(hl_cfg_p H,int line){
  array_index_t k;
  
  k=hl_head(H,line);
  
  do {
    hl_cell_p p=H->cell+k;
    fprintf(stderr, "%c%d[%d]@%d\t",
	    sign2char(p->real_sign), 
	    p->mu,p->bindex,k);
    k=hl_rnext(H,k);
  } while (k!=hl_head(H,line));

  fprintf(stderr,"\n");
}

  

extern void hl_dump(hl_cfg_p H){
  int i,j;

  for (j=0; j< sa_size(H->clist); j++){
    array_index_t k=sa_read(H->clist,j);

    if (k!=-1){
      fprintf(stderr,"%d,%d\t",j,k);
    }
  }
  fprintf(stderr,"\n");
  for (i=0; i< H->rows; i++){

    hl_cell_p p=H->cell+hl_head(H,i);
    fprintf(stderr,"[%d%c%d] ",
	    i,
	    (sa_read(H->ridge,H->row_hlindex[i])) > 0 ? 'r' : ' ',
	    H->row_hlindex[i]);
    
    for (j=0; j< H->cols; j++){
      fprintf(stderr,"%d,%c%c%c%d(%d),%d,%d\t", hl_head(H,i)+j,
	      sign2char(p->tau),
	      sign2char(p->real_sign),
	      sign2char(p->sign),
	      p->mu, p->bindex,
	      p->rprev,p->rnext);
      p++;
    }

    fprintf(stderr,"\n");
  }
  fprintf(stderr,"\n");
}

	     


#ifdef HL_TEST
#define SLOT_hyp 0
#define SLOT_chirotope SLOT_hyp+WIDTH_hl

extern void hl_test(array_index_t n,array_index_t r,sarray_p chirotope){
  na_node_p node;
  hl_cfg_p H;


  n=5;
  r=3;

  node=na_new();
  node=hl_nalloc(node,SLOT_hyp,n,r);

  H=na_access(node,SLOT_hyp);


  hl_init(H,n,r,sa_read(chirotope,0));
  hl_dump(H);
  try_both_signs(H,chirotope,r+1);

}
#endif

/** 
 * update feasible interval on hyperline rows according to a basis sign
 * 
 * @param local search state
 * @param global global tables
 * @param chirotope 
 * @param H hyperline arrangement
 * @param table_beta basis index as it appears in H
 * @param sign sign that the basis is being set to
 * 
 * @return 1 if successful
 */
extern bool hl_set_gaps(na_node_p local, na_node_p global,
			sarray_p chirotope,
			hl_cfg_p H, 
			array_index_t table_beta,
			sign_t sign){

  array_index_t cell_idx;
  int num_occurances,i;
  array_index_t * occurance;
  array_index_t beta0;
  sign_t sigma0;

  /*
  if (table_beta==11)
    return 1;
  */
  oms_debug(3,("hl_set_gaps","%d %d",table_beta,sign));
  DEBUG_TAB;
    /* the occurances must be treated from top of table. Possibly a
       change of data structure is in order, since the current one
       stores the from the bottom up */

    num_occurances=binomial(H->rank-1,H->rank-2);
    occurance=calloc(num_occurances,sizeof(array_index_t));

   i=num_occurances;

    cell_idx=sa_read(H->clist,table_beta);
    while(cell_idx>=0){
      occurance[--i]=cell_idx;
      cell_idx=H->cell[cell_idx].cnext;
    }
    for ( /* use value above */; i< num_occurances; i++){

      hl_cell_p cell;

      cell_idx=occurance[i];
      cell=H->cell+cell_idx;


      beta0=hl_index(H,hl_head(H,cell->row));
      sigma0=sa_read(chirotope,beta0);
      if (sigma0==0){
	  /* Do Nothing, this row of the table is not active yet */
      } else {	  

	/* is it ok to insert the new element after 'cell' */
	
	array_index_t old_left,old_right;
	old_left=hl_left(H,cell->row);
	old_right=hl_right(H,cell->row);

	if (hl_key(H,cell_idx) <= hl_key(H,old_right) &&
	    sign*hl_new_real_sign(H,cell->row,chirotope)
	    *(cell->real_sign)*cell->tau!=1){ 
	    array_index_t k;
	    
	    /* Right gap bound changed */
	    
	    if (hl_key(H,cell_idx) < hl_key(H,old_left)){
	      free(occurance);
	      return FALSE;
	    } else {

	      hl_set_right(H,cell->row,cell->rprev);
	      oms_debug(4,("hl_set_gaps","right %d %d->%d",cell->row,old_right,
			   hl_right(H,cell->row)));
  
	      for (k=cell->rnext; 
		   hl_key(H,cell_idx)< hl_key(H,k) &&hl_key(H,k) <= hl_key(H,old_right);
		   k=hl_rnext(H,k)){
		/* these are the cocircuits we skipped over */
		hl_cell_p ocell=H->cell+k;
		
		int new_sign=hl_new_sign(H,ocell->row,chirotope)*(ocell->sign)*-1;
		array_index_t new_bindex=hl_index(H,k);
	    
		if (!SetSign(local,global,new_bindex,new_sign)){
		  free(occurance);
		  RETURN( FALSE);
		}
	      }
	    } 
	}
	/* is it ok to insert the new element before 'cell' */
	/** \todo is the second condition correct? */
	if (hl_key(H,cell_idx) > hl_key(H,hl_left(H,cell->row)) &&
	    sign*hl_new_real_sign(H,cell->row,chirotope)
	    *(cell->real_sign)*cell->tau!=-1){ 
	    array_index_t k;

	    /* Left gap bound changed */

	    if (hl_key(H,cell_idx) > hl_key(H,old_right)){
	      free (occurance);
	      return FALSE;
	    } else {

	      hl_set_left(H,cell->row,cell_idx);
	      
	      
	      oms_debug(4,("hl_set_gaps","left %d %d->%d",cell->row,old_left,
			 hl_left(H,cell->row)));

	      for ( k=hl_rnext(H,old_left); 
		    hl_key(H,old_left)< hl_key(H,k) &&hl_key(H,k) < hl_key(H,cell_idx);
		    k=hl_rnext(H,k) )  
		{ 
		  hl_cell_p ocell=H->cell+k;
		
		  int new_sign  = hl_new_sign(H,ocell->row,chirotope) * ocell->sign;
		  int new_bindex = hl_index(H,k);

		  if ( !SetSign(local,global, new_bindex, new_sign) ) {
		    free(occurance);
		    RETURN(FALSE);
		  }
		}
	    }
	    
	}

      } 
    }
    free(occurance);
    RETURN(TRUE);
}

/* check that the hyperline  at position %$\rho$% is non-alternating. */
bool hl_check_alternation(hl_cfg_p H, array_index_t row){

  int k;
  int switches=0;
  int last_sign=1;

  /* note: the first real_sign in a row should always be positive */

  k=hl_head(H,row);
  do { 
    if (H->cell[k].real_sign != last_sign)
      switches++;
    if (switches>1) 
      return FALSE;
    last_sign=H->cell[k].real_sign;
    k=hl_rnext(H,k);
  } while (k!=hl_head(H,row));
  return TRUE;
}
	
/*
  Mark all of the hyperlines contained in the basis b as 
  being ridges, i.e. on the boundary.
*/

bool hl_set_ridges(hl_cfg_p H, array_index_t b, be_block_p ridge_table){
  int i;

  for (i=0; i< be_cols(ridge_table); i++){
    array_index_t hl_index=be_bindex(ridge_table,b,i);
    array_index_t hl_row=H->hlindex_row[hl_index];

    if (hl_row != -1 &&  !hl_check_alternation(H,hl_row)) 
      return FALSE;

    sa_set(H->ridge,hl_index,1);

  }
  return TRUE;
}

/** 
 * Build a table from basis indices to those stored in the hyperline
 * arrangment.
 * 
 * @param t table
 * @param n number of elements
 * @param r rank
 */
static void hl_init_bindex_table(int *t, int n, int r){
  
  natural_t *pi,*pibar;
  sign_t *tau;
  array_index_t i;

  pi=calloc(r,sizeof(natural_t));
  pibar=calloc(n-r,sizeof(natural_t));
  tau=calloc(n-r,sizeof(sign_t));

  first_kpartition(n,r,pi,pibar,tau);

  i=0;

  do {
    natural_t old_max;
    
    old_max=pi[r-1];
    pi[r-1]=n;
    t[i]=kset_index(n,r,pi);
    pi[r-1]=old_max;
    i++;
  } while (next_kpartition(n,r,pi,pibar,tau));

  free(pi);
  free(pibar);
  free(tau);
}
